/*  $Id: Crypto.java 1335 2012-08-25 13:27:49Z sdevaux $
 *
 * Tabs = 4 chars
 * copyright (C) 2002-2012 Sebastien Devaux
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 *              Sebastien Devaux <sebastien.devaux@laposte.net>
 */
package net.eduvax.util;

import java.util.Random;
import java.nio.charset.Charset;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
/** 
 * Simple crypto utility.  
 */ 
public class Crypto
{
    /**
     * Create new Crypto tool.
     * @param key cipher key.
     */ 
    public Crypto(String key) {
        _key=key.getBytes(Charset.forName("UTF-8"));
        if (_md5==null) {
                try {
                    _md5=MessageDigest.getInstance("MD5");
                } catch (NoSuchAlgorithmException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
        }
    }
    /** 
     * Reversible crypt of a string.
     * The ciphered string can e deciphered using the uncrypt()
     * method. The encryption is non determinist, even when using
     * the same encryption key, a random seed will produce a different
     * encryption at each use.
     * @param str string to crypt.
     * @return cyphered string.
     */ 
    public String reversibleCrypt(String str) 
    {
        if (str == null || str.trim() == "")
        {
            return "";
        }

        return cypher(str,false);
    }

    /** 
     * Non reversible encryption.
     * This encryption is base on hash code. This encryption method
     * is determinist. Using the same key, a string will always produce
     * the same encrypted string. Then to check validity of a candidate
     * password, just encrypt the candidate to compare with the stored
     * encrypted string.
     * @param str string to be encrypted.
     * @return encrypted string.
     */ 
    public String nonReversCrypt(String str) 
    {
        if (str == null || str.trim() == "")
        {
            return "";
        }
        String cryptage=cypher(str,true);
        // hashcode MD5 du la chaîne cryptée...
        return getMD5HashString(cryptage);
    }

    /** 
     * Compute MD5 hash string.
     * @param str string to hash
     * @return string representation of the hash code.
     */ 
    static public String getMD5HashString(String str) 
    {
        byte[] bytes=str.getBytes(Charset.forName("UTF-8"));
        if (_md5!=null) {
            bytes=_md5.digest(bytes);
        }
        StringBuffer result=new StringBuffer();
        for (int i=0;i<bytes.length;i++) 
        {
            result.append(String.format("%x",bytes[i]));
        }
        return result.toString();
    }
    /** 
     * Decrypt a string encrypted in a revesible way.
     * This method is dedicated to decrypt string encrypted using 
     * the reversibleCrypt() method.
     * @param str cyphered string.
     * @return decyphered string.
     */ 
    public String uncrypt(String str) 
    {
        if (str == null || str.trim() == "")
        {
            return "";
        }

        byte[] cryptage=base64Decode(str);
        byte[] clair=new byte[cryptage.length];
        byte[] random={cryptage[0],cryptage[1]};
        for (int i=0;i<cryptage.length-2;i++) 
        {
            byte ch=(byte)(cryptage[i+2]^
                            (_key[i%_key.length]+random[(i)%2]));
            if (ch==0)
            {
                return new String(clair,0,i);
            }
            else
            {
                clair[i]=ch;
            }
        }
        return new String(clair);
    }
    /** 
     * Simple crypto implementation.
     * input string XOR (cypher key + optional random seed). The cyhered
     * bytes are base 64 encoded to keep encryption compatible with ASCII
     * streams. 
     * @param str string to crypt
     * @param determinist crypt methode : with random seed (true), without
     * random seed (false)
     * @return crypted string.
     */		 
    private String cypher(String str,boolean determinist) 
    {
        byte[] input=str.getBytes();
        int longueur = input.length / 10;
        longueur++;
        longueur*=10;
        byte[] random={0,0};
        if (determinist)
        {
            short r=(short)(input.length>0?input[0]:0
                        +(input.length>1?input[1]:0<<8)+str.length());
            random[0]=(byte)(r&0xFF);
            random[1]=(byte)(r>>8&0xFF);
        }
        else
        {
            _rnd.nextBytes(random);
        }
        byte[] cryptage=new byte[longueur+2];

        int i=0;
        cryptage[0]=random[0];
        cryptage[1]=random[1];
        for (;i<input.length;i++) 
        {
            cryptage[i+2]=(byte)(input[i]^(_key[i%_key.length]+random[i%2]));
        }
        i+=2;
        for (;i<cryptage.length;i++) 
        {
            cryptage[i]=(byte)(0^(_key[(i-2)%_key.length]+random[(i-2)%2]));			
        }
        return base64Encode(cryptage);
    }
    /** 
     * Simple base 64 encoding.
     * Not real full base 64 encoding, produce string without any CR char.
     * Regular 72 char line split is not produced.
     * @param data data to be encoded.
     * @return base 64 encoded string.
     */ 
    static public String base64Encode(byte[] data) 
    {
        int padding=(3-(data.length%3))%3;
        StringBuffer buf=new StringBuffer(3*data.length/2+padding);
        for (int i=0;i<data.length;)
        {
            byte a=data[i];
            i++;
            byte b=i<data.length?data[i]:0;
            i++;
            byte c=i<data.length?data[i]:0;
            i++;

            int j = ((a>=0?a:256+a) << 16) + ((b>=0?b:256+b) << 8) + (c>=0?c:256+c);
            buf.append(_base64.charAt((j >> 18) & 0x3f));
            buf.append(_base64.charAt((j >> 12) & 0x3f));
            buf.append(_base64.charAt((j >> 6) & 0x3f));
            buf.append(_base64.charAt(j & 0x3f));
        }
        switch (padding) {
            case 1:
                buf.append('=');
                break;
            case 2:
                buf.append("==");
                break;
        }
        return buf.toString();
    }
    /** 
     * Simple base 64 decoding.
     * Dedicated to base 64 string created by this class's base64Encoding 
     * method. Does not handle some regular base 64 encoding such as
     * multiline encoding.
     * @param str string to decode.
     * @return decoded data as byte array.
     */ 
    static public byte[] base64Decode(String str) 
    {
        int inLen=str.length();
        int padding=0;
        if (str.charAt(inLen-1)=='=') {
            inLen--;
            padding++;
            if (str.charAt(inLen-1)=='=') {
                inLen--;
                padding++;
            }
        }
        byte[] buf=new byte[inLen*3/4-padding];
        int j=0;
        for (int i=0;i<buf.length;) {
            int c=_base64.indexOf(str.charAt(j))<<18;
            j++;
            c+=_base64.indexOf(str.charAt(j))<<12;
            j++;
            c+=_base64.indexOf(str.charAt(j))<<6;
            j++;
            c+=_base64.indexOf(str.charAt(j));
            j++;
            buf[i]=(byte)((c&0xff0000)>>16);
            i++;
            if (i<buf.length) {
                buf[i]=(byte)((c&0xff00)>>8);
                i++;
                if (i<buf.length) {
                    buf[i]=(byte)(c&0xff);
                    i++;
                }
            }
        }
        return buf;
    }


    /** 
     * Cypher key.
     */
    private byte[] _key={'a','b','c','d','e','f'};
    /** 
     * base 64 code set.
     */
    static private String _base64=
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
    /** 
     * Random generator.
     * @djin.relation composition
     */ 
    static private Random _rnd = new Random();
    /** 
     * Hash code provider.
     * @djin.relation composition
     */ 
    static private MessageDigest _md5 = null;
}
